// ==========================================================================
// Dedmonwakeen's Raid DPS/TPS Simulator.
// Send questions to natehieter@gmail.com
// ==========================================================================

#include "simulationcraft.hpp"

std::string progress_bar_t::format_time( double t )
{
  std::stringstream s;

  int days = 0, hours = 0, minutes = 0;

  double remainder = t;

  if ( remainder >= 86400 )
  {
    days = static_cast<int>(remainder / 86400);
    remainder -= days * 86400;
  }

  if ( remainder >= 3600 )
  {
    hours = static_cast<int>(remainder / 3600);
    remainder -= hours * 3600;
  }

  if ( remainder >= 60 )
  {
    minutes = static_cast<int>(remainder / 60);
    remainder -= minutes * 60;
  }

  if ( days > 0 )
  {
    s << days << "d, ";
  }

  if ( hours > 0 )
  {
    s << hours << "h, ";
  }

  if ( minutes > 0 )
  {
    s << minutes << "m, ";
  }

  s << util::round( remainder, 0 ) << "s";

  return s.str();
}

progress_bar_t::progress_bar_t( sim_t& s ) :
    sim( s ), steps( 20 ), updates( 100 ), interval( 0 ), update_number( 0 ),
    start_time( 0 ), last_update( 0 ), max_interval_time( 1.0 ), work_index( 0 ), total_work_( 0 ),
    elapsed_time( 0 ), time_count( 0 )
{
}

void progress_bar_t::init()
{
  start_time = util::wall_time();
  if ( sim.target_error > 0 )
  {
    interval = sim.analyze_error_interval;
  }
  else
  {
    interval = sim.work_queue -> size();
    if ( sim.deterministic || sim.strict_work_queue )
    {
      interval *= sim.threads;
    }
    interval /= updates;
  }
  if ( interval == 0 )
  {
    interval = 1;
  }
}

void progress_bar_t::restart()
{
  add_simulation_time( util::wall_time() - start_time );

  start_time = util::wall_time();
  last_update = util::wall_time();
  update_number = 0;
}

bool progress_bar_t::update( bool finished, int index )
{
  if ( sim.thread_index != 0 )
  {
    return false;
  }

  if ( ! sim.report_progress )
  {
    return false;
  }

  if ( ! sim.current_iteration )
  {
    return false;
  }

  if ( ! finished && sim.target_error > 0 && sim.current_iteration < sim.analyze_error_interval )
  {
    return false;
  }

  auto progress = sim.progress( nullptr, index );
  if ( ! finished )
  {
    double update_interval = last_update - util::wall_time();
    if ( progress.current_iterations < update_number + interval &&
         update_interval < max_interval_time )
    {
      return false;
    }
    update_number = progress.current_iterations;
    last_update = util::wall_time();
  }

  if ( sim.progressbar_type == 1 )
  {
    return update_simple( progress, finished, index );
  }
  else
  {
    return update_normal( progress, finished, index );
  }
}

bool progress_bar_t::update_simple( const sim_progress_t& progress, bool finished, int /* index */ )
{
  auto pct = progress.pct();
  if ( pct <= 0 )
  {
    return false;
  }

  if ( finished )
  {
    pct = 1.0;
  }

  double current_time = util::wall_time() - start_time;
  double total_time = current_time / pct;

  double remaining_time = total_time - current_time;

  status = util::to_string( finished ? progress.total_iterations : progress.current_iterations );
  status += '\t';
  status += util::to_string( progress.total_iterations );
  if ( sim.target_error > 0 )
  {
    status += '\t';
    status += util::to_string( sim.current_mean );
    status += '\t';
    status += util::to_string( sim.current_error );
  }

  if ( remaining_time > 0 )
  {
    status += '\t';
    status += util::to_string( util::round( remaining_time, 3 ) );
  }
  else
  {
    if ( finished && current_time > 0 )
    {
      status += '\t';
      status += util::to_string( util::round( current_time, 3 ) );
    }
    else
    {
      status += "\t0.000";
    }
  }

  if ( ! finished && total_work() > 0 )
  {
    auto average_spent = average_simulation_time();
    auto phases_left = total_work() - current_progress();
    auto time_left = std::max( 0.0, average_spent - ( util::wall_time() - start_time ) );
    auto total_left = phases_left * average_spent + time_left;

    if ( total_left > 0 )
    {
      status += "\t";
      status += format_time( total_left );
    }
  }

  return true;
}

bool progress_bar_t::update_normal( const sim_progress_t& progress, bool finished, int /* index */ )
{
  auto pct = progress.pct();
  if ( pct <= 0 )
  {
    return false;
  }

  if ( finished )
  {
    pct = 1.0;
  }

  size_t prev_size = status.size();

  status = "[";
  status.insert( 1, steps, '.' );
  status += "]";

  int length = static_cast<int>( steps * pct + 0.5 );
  for ( int i = 1; i < length + 1; ++i )
  {
    status[ i ] = '=';
  }

  if ( length > 0 )
  {
    status[ length ] = '>';
  }

  double current_time = util::wall_time() - start_time;
  double total_time = current_time / pct;

  int remaining_sec = static_cast<int>( total_time - current_time );
  int remaining_min = remaining_sec / 60;
  remaining_sec -= remaining_min * 60;

  str::format( status, " %d/%d", finished ? progress.total_iterations : progress.current_iterations, progress.total_iterations );

  if ( sim.target_error > 0 )
  {
    str::format( status, " Mean=%.0f Error=%.3f%%", sim.current_mean, sim.current_error );
  }

  if ( remaining_min > 0 )
  {
    str::format( status, " %dmin", remaining_min );
  }

  if ( remaining_sec > 0 )
  {
    str::format( status, " %dsec", remaining_sec );
  }

  if ( finished )
  {
    int total_min = current_time / 60;
    int total_sec = fmod( current_time, 60 );
    int total_msec = 1000 * ( current_time - static_cast<int>( current_time ) );
    if ( total_min > 0 )
    {
      str::format( status, " %dmin", total_min );
    }

    if ( total_sec > 0 )
    {
      str::format( status, " %d", total_sec );
      if ( total_msec > 0 )
      {
        str::format( status, ".%d", total_msec );
      }

      status += "sec";
    }
    else if ( total_msec > 0 )
    {
      str::format( status, " %dmsec", total_msec );
    }
  }

  if ( total_work() > 0 )
  {
    auto average_spent = average_simulation_time();
    auto phases_left = total_work() - current_progress();
    auto time_left = std::max( 0.0, average_spent - ( util::wall_time() - start_time ) );
    auto total_left = phases_left * average_spent + time_left;

    if ( total_left > 0 )
    {
      status += " (";
      status += format_time( total_left );
      status += ")";
    }
  }

  if ( prev_size > status.size() )
  {
    status.insert( status.end(), ( prev_size - status.size() ), ' ' );
  }

  return true;
}

void progress_bar_t::output( bool finished )
{
  if ( ! sim.report_progress )
  {
    return;
  }

  char delim = sim.progressbar_type == 1 ? '\t' : ' ';
  char terminator = ( sim.progressbar_type == 1 || finished ) ? '\n' : '\r';
  std::stringstream s;
  if ( sim.progressbar_type == 0 )
  {
    s << "Generating ";
  }

  s << base_str;
  // Separate base and phase by a colon for easier parsing
  if ( sim.progressbar_type == 0 )
  {
    s << ":";
  }

  if ( ! phase_str.empty() )
  {
    s << delim;
    s << phase_str;
  }

  s << delim;
  s << current_progress();
  if ( sim.progressbar_type == 0 )
  {
    s << "/";
  }
  else
  {
    s << "\t";
  }
  s << compute_total_phases();
  s << delim;
  s << status;
  s << terminator;

  std::cout << s.str();
  fflush( stdout );
}

void progress_bar_t::progress()
{
  if ( sim.thread_index > 0 )
  {
    return;
  }

  if ( sim.parent )
  {
    sim.parent -> progress_bar.progress();
  }
  else
  {
    work_index++;
  }
}

size_t progress_bar_t::compute_total_phases()
{
  if ( sim.parent )
  {
    return sim.parent -> progress_bar.compute_total_phases();
  }

  if ( total_work_ > 0 )
  {
    return total_work_;
  }

  size_t n_actors = 1;
  if ( sim.single_actor_batch )
  {
    n_actors = sim.player_no_pet_list.size();
  }

//  size_t reforge_plot_phases = 0;
//  if ( sim.reforge_plot -> num_stat_combos > 0 )
//  {
//    reforge_plot_phases = n_actors * sim.reforge_plot -> num_stat_combos;
//  }

  auto work = n_actors /* baseline */ +
              n_scale_factor_phases() +
              n_plot_phases() +
              n_reforge_plot_phases() +
              sim.profilesets.n_profilesets();

  // Once profilesets have all been constructed, cache the total work done since we can determine
  // the maximum number of phases then
  if ( sim.profilesets.is_running() )
  {
    total_work_ = work;
  }

  return total_work_ > 0 ? total_work_ : work;
}

void progress_bar_t::set_base( const std::string& base )
{
  base_str = base;
}

void progress_bar_t::set_phase( const std::string& phase )
{
  phase_str = phase;
}

size_t progress_bar_t::current_progress() const
{
  if ( sim.parent )
  {
    return sim.parent -> progress_bar.current_progress();
  }

  return work_index;
}

size_t progress_bar_t::total_work() const
{
  if ( sim.parent )
  {
    return sim.parent -> progress_bar.total_work();
  }

  return total_work_;
}

size_t progress_bar_t::n_stat_scaling_players( const std::string& stat_str ) const
{
  auto stat = util::parse_stat_type( stat_str );
  if ( stat == STAT_NONE )
  {
    return 0;
  }

  // Ensure sim has at least someone who scales with the stat
  return std::count_if( sim.player_no_pet_list.begin(), sim.player_no_pet_list.end(),
    [ stat ]( const player_t* p ) {
      return ! p -> quiet && p -> scaling -> scales_with[ stat ];
    } );
}

size_t progress_bar_t::n_plot_phases() const
{
  if ( sim.plot -> dps_plot_stat_str.empty() )
  {
    return 0;
  }

  size_t n_phases = 0;
  auto stat_list = util::string_split( sim.plot -> dps_plot_stat_str, ",:;/|" );
  range::for_each( stat_list, [ &n_phases, this ]( const std::string& stat_str ) {
    auto n_players = n_stat_scaling_players( stat_str );

    if ( n_players > 0 )
    {
      // Don't use fancy context-sensitive number of phases, but rather just multiply with the
      // number of actors if single_actor_batch=1. In the future if scale factor calculation is made
      // context (stat) sensitive, adjust this.
      /* n_phases += sim.plot -> dps_plot_points * ( sim.single_actor_batch == 1 ? n_players : 1 ); */
      n_phases += sim.plot -> dps_plot_points *
                  ( sim.single_actor_batch == 1 ? sim.player_no_pet_list.size() : 1 );
    }
  } );

  return n_phases;
}

size_t progress_bar_t::n_scale_factor_phases() const
{
  if ( ! sim.scaling -> calculate_scale_factors )
  {
    return 0;
  }

  size_t n_phases = 0;
  auto stat_list = util::string_split( sim.scaling -> scale_only_str, ",:;/|" );
  std::vector<stat_e> scale_only;

  range::for_each( stat_list, [ &scale_only ]( const std::string& stat_str ) {
    auto stat = util::parse_stat_type( stat_str );
    if ( stat != STAT_NONE )
    {
      scale_only.push_back( stat );
    }
  } );

  for ( stat_e stat = STAT_NONE; stat < STAT_MAX; ++stat )
  {
    if ( scale_only.size() > 0 && range::find( scale_only, stat ) == scale_only.end() )
    {
      continue;
    }

    auto n_players = std::count_if( sim.player_no_pet_list.begin(), sim.player_no_pet_list.end(),
      [ stat ]( const player_t* p ) {
        return ! p -> quiet && p -> scale_player && p -> scaling -> scales_with[ stat ];
      } );

    if ( n_players > 0 )
    {
      // Don't use fancy context-sensitive number of phases, but rather just multiply with the
      // number of actors if single_actor_batch=1. In the future if scale factor calculation is made
      // context (stat) sensitive, adjust this.
      /* n_phases += sim.single_actor_batch == 1 ? n_players : 1; */
      n_phases += sim.single_actor_batch == 1 ? sim.player_no_pet_list.size() : 1;

      // Center delta doubles the number of scale factor phases
      if ( sim.scaling -> center_scale_delta )
      {
        n_phases += sim.single_actor_batch == 1 ? sim.player_no_pet_list.size() : 1;
      }
    }
  }

  return n_phases;
}

size_t progress_bar_t::n_reforge_plot_phases() const
{
  if ( sim.reforge_plot -> reforge_plot_stat_str.empty() )
  {
    return 0;
  }

  auto stat_list = util::string_split( sim.reforge_plot -> reforge_plot_stat_str, ",:;/|" );
  std::vector<stat_e> stat_indices;
  range::for_each( stat_list, [ &stat_indices, this ]( const std::string& stat_str ) {
    if ( n_stat_scaling_players( stat_str ) > 0 )
    {
      stat_indices.push_back( util::parse_stat_type( stat_str ) );
    }
  } );

  std::vector<int> cur_stat_mods( stat_indices.size() );
  std::vector<std::vector<int>> stat_mods;

  sim.reforge_plot -> generate_stat_mods( stat_mods, stat_indices, 0, cur_stat_mods );

  return sim.single_actor_batch == 1
         ? sim.player_no_pet_list.size() * stat_mods.size()
         : stat_mods.size();
}

void progress_bar_t::add_simulation_time( double t )
{
  if ( t <= 0 )
  {
    return;
  }

  if ( sim.parent )
  {
    sim.parent -> progress_bar.add_simulation_time( t );
  }
  else
  {
    elapsed_time += t;
    time_count++;
  }
}

double progress_bar_t::average_simulation_time() const
{
  if ( sim.parent )
  {
    return sim.parent -> progress_bar.average_simulation_time();
  }

  return time_count > 0 ? elapsed_time / time_count : 0;
}
